//------------------------------------------------------------------------------
// File: CD3DFilter.cpp
//
// Desc: This class utilizes a hidden D3D window that it renders to in the
//		 process of transforming each recieved frame. One can change the
//		 parameters that are used in the pixel shader through the interface.
//		 The pixel shader is compiled each time the filter is run, so it can
//		 be changed between the runs (to make different effects).
//
//------------------------------------------------------------------------------
//
//


#include <windows.h>
#include <streams.h>
#include <initguid.h>
#include <olectl.h>
#include <math.h>

#include "dxutil.h"

#if (1100 > _MSC_VER)
#include <olectlid.h>
#endif
#include "d3dfilteruids.h"
#include "id3dfilter.h"
#include "d3dfilterprop.h"
#include "cd3dfilter.h"


void Msg(TCHAR *szFormat, ...);
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

// for logging
//using namespace std;

#pragma warning(disable:4238)  // nonstd extension used: class rvalue used as lvalue

// setup data

WNDCLASSEX wc;

// A structure for our custom vertex type. We added texture coordinates
struct CUSTOMVERTEX
{
    FLOAT  x, y, z; // The position
    DWORD  color;    // The color
    FLOAT  tu, tv;   // The texture coordinates
};

// Our custom FVF, which describes our custom vertex structure
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX1)


const AMOVIESETUP_MEDIATYPE sudPinTypes =
{
    &MEDIATYPE_Video,       // Major type
    &MEDIASUBTYPE_NULL      // Minor type
};

const AMOVIESETUP_PIN psudPins[] =
{
    {
        L"Input",           // String pin name
        FALSE,              // Is it rendered
        FALSE,              // Is it an output
        FALSE,              // Allowed none
        FALSE,              // Allowed many
        &CLSID_NULL,        // Connects to filter
        L"Output",          // Connects to pin
        1,                  // Number of types
        &sudPinTypes },     // The pin details
      { L"Output",          // String pin name
        FALSE,              // Is it rendered
        TRUE,               // Is it an output
        FALSE,              // Allowed none
        FALSE,              // Allowed many
        &CLSID_NULL,        // Connects to filter
        L"Input",           // Connects to pin
        1,                  // Number of types
        &sudPinTypes        // The pin details
    }
};


const AMOVIESETUP_FILTER sudContrast =
{
    &CLSID_D3DFilter,        // Filter CLSID
    L"D3D Video Filter",      // Filter name
    MERIT_DO_NOT_USE,       // Its merit
    2,                      // Number of pins
    psudPins                // Pin details
};


// List of class IDs and creator functions for the class factory. This
// provides the link between the OLE entry point in the DLL and an object
// being created. The class factory will call the static CreateInstance

CFactoryTemplate g_Templates[2] = {

    { L"D3D Video Filter"
    , &CLSID_D3DFilter
    , CD3DFilter::CreateInstance
    , NULL
    , &sudContrast }
  ,
    { L"Video Contrast Property Page"
    , &CLSID_D3DFilterPropertyPage
    , CD3DFilterProperties::CreateInstance }
};
int g_cTemplates = sizeof(g_Templates) / sizeof(g_Templates[0]);


//
// Constructor
//
CD3DFilter::CD3DFilter(TCHAR *tszName,LPUNKNOWN punk,HRESULT *phr) :
    CTransformFilter(tszName, punk, CLSID_D3DFilter),
	m_lVidWidth(0),
	m_lVidHeight(0),
	m_lVidPitch(0),
	m_lBufferRequest(0),
	m_pTexture(NULL),
	m_pRenderTargetSurface(NULL),
	m_pRenderToSurface(NULL)
{
	params[0] = 0.5f;
	params[1] = 0.5f;
	params[2] = 0.5f;
	params[3] = 0.5f;

	//params = new float[4];
    ASSERT(tszName);
    ASSERT(phr);

	// create logfile
	m_logfile = new ofstream(LOGFILE_LOCATION);

	*m_logfile << "%TIME = [ulTex render dlTex];" << endl;
	m_logfile->flush();


	//CREATE WINDOW
    // Register the window class
    WNDCLASSEX i_wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "D3D Filter", NULL };
	m_wc = wc = i_wc;

	if(!SUCCEEDED(RegisterClassEx( &m_wc )))
		*m_logfile << LOCATION << "register failed";
 
	// Create the application's window
	 hWnd = CreateWindow( "D3D Filter", "D3D Filter: Render window",
                              WS_OVERLAPPED, 100, 100, 300, 300,
                              GetDesktopWindow(), NULL, NULL, NULL );
	 

	if(hWnd == NULL)
		*m_logfile << LOCATION << "hWnd is NULL. Error is: " << GetLastError() << ".\n";

	// INITIALIZE D3D device
	if( FAILED( this->InitD3D(hWnd) ) )
		*m_logfile << LOCATION << "initializing of d3d failed";	
	
	// Create the vertex buffer.
	if( FAILED( m_pd3dDevice->CreateVertexBuffer( 4*sizeof(CUSTOMVERTEX),
                                      0, D3DFVF_CUSTOMVERTEX,
                                      D3DPOOL_DEFAULT, &m_pVB, NULL ) ) )
    {
        Msg(TEXT("Could not create a vertex buffer!"));
    }

    // Fill the vertex buffer. We are setting the tu and tv texture
    // coordinates, which range from 0.0 to 1.0

	
	CUSTOMVERTEX pVertices[]=
	{
		//  x      y     z     diffuse	  tu tv 
		{ -1.0f, -1.0f, 0.0f, 0xffff0000, 0, 1 },	//  lower right - red
		{ -1.0f,  1.0f, 0.0f, 0xff00ff00, 0, 0 },	//  lower left - green
		{  1.0f, -1.0f, 0.0f, 0xff0000ff, 1, 1 },	//  upper left - blue
		{  1.0f,  1.0f, 0.0f, 0xffffffff, 1, 0 },	//  upper right - white
	};

	void* gVertices;
    if ( FAILED(m_pVB->Lock( 0, sizeof(pVertices), (void**)&gVertices, 0 ) ) )
    {
        Msg(TEXT("Could not lock the vertex buffer!"));
    }
	memcpy(gVertices,pVertices,sizeof(pVertices));
	m_pVB->Unlock();
    /*
	
	CUSTOMVERTEX* pVertices;
    if ( FAILED(m_pVB->Lock( 0, 0, (VOID**)&pVertices, 0 ) ) )
    {
        Msg(TEXT("Could not lock the vertex buffer!"));
    }
	
	pVertices[0].position = D3DXVECTOR3( -1.0f, -1.0f, 0.0f );
    pVertices[0].color    = 0xffffffff;
    pVertices[0].tu       = 0;
    pVertices[0].tv       = 1;

    pVertices[1].position = D3DXVECTOR3( -1.0f, 1.0f, 0.0f );
    pVertices[1].color    = 0xffffffff;
    pVertices[1].tu       = 0;
    pVertices[1].tv       = 0;

    pVertices[2].position = D3DXVECTOR3( 1.0f, -1.0f, 0.0f );
    pVertices[2].color    = 0xffffffff;
    pVertices[2].tu       = 1;
    pVertices[2].tv       = 1;

    pVertices[3].position = D3DXVECTOR3( 1.0f, 1.0f, 0.0f );
    pVertices[3].color    = 0xffffffff;
    pVertices[3].tu       = 1;
    pVertices[3].tv       = 0;

    m_pVB->Unlock();
*/

	HRESULT hr;

	if(FAILED(hr=m_pd3dDevice->CreateRenderTarget(640, 480, D3DFMT_A8R8G8B8, D3DMULTISAMPLE_NONE, 0, FALSE, &m_pRenderTargetSurface, NULL)) )
	//if(FAILED(hr=D3DXCreateTexture(m_pd3dDevice, 640, 480, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_pRenderTargetSurface)) )
	{
		*m_logfile << LOCATION << "Error when creating render target texture" << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();
	}


	D3DVERTEXELEMENT9 decl[] = 
	{
		{ 0, 0,  D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 }, // pos
		{ 0, 0,  D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0 }, // col
		{ 0, 0,  D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_UV, D3DDECLUSAGE_TEXCOORD, 1 }, // tex
		D3DDECL_END()
	};

    // Create the vertex shaders.
    TCHAR        strPath[512];                      // location of the shader file
    LPD3DXBUFFER pCode;                             // assembled shader code
	LPD3DXBUFFER pErrorMsgs;						// Error messages
	lstrcpy( strPath, VERTEX_SHADER_LOCATION );

	if( FAILED(hr = D3DXAssembleShaderFromFile( strPath, NULL, NULL, D3DXSHADER_DEBUG, &pCode, &pErrorMsgs )) ) {
		if(hr==D3DERR_INVALIDCALL)
			Msg(TEXT("Assemble vertex shader from file FAILED: D3DERR_INVALIDCALL"));
		else if(hr==D3DXERR_INVALIDDATA)
			Msg(TEXT("Assemble vertex shader from file FAILED: D3DXERR_INVALIDDATA"));
		else if(hr==E_OUTOFMEMORY)
			Msg(TEXT("Assemble vertex shader from file FAILED: E_OUTOFMEMORY"));
		else
			Msg(TEXT("Assemble vertex shader from file FAILED: This error is not possible..."));

		*m_logfile << LOCATION << "did not find vertex shader in" << strPath << endl;	
		*m_logfile << "Error is: " << (hr&0xffff) << endl;
		m_logfile->flush();
	}
	else
	{
		m_pd3dDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &m_hBasicVertexShader );
	}
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	// Create the pixel shader
	lstrcpy( strPath, PIXEL_SHADER_LOCATION );
	if( FAILED(D3DXAssembleShaderFromFile( strPath, NULL, NULL, 0, &pCode, &pErrorMsgs )) ) {  // assemble shader code
		Msg(TEXT("Assemble pixel shader from file FAILED (see logfile)") );// assemble shader code
		*m_logfile << LOCATION << "did not find vertex shader in" << strPath;	
		m_logfile->flush();
	}
	else {
		m_pd3dDevice->CreatePixelShader( (DWORD*)pCode->GetBufferPointer(), &m_hBasicPixelShader );
	}	
	if(pCode!=NULL)
		pCode->Release();
	if(pErrorMsgs!=NULL)
		pErrorMsgs->Release();

	m_logfile->flush();

} // Contrast

CD3DFilter::~CD3DFilter() {

	delete m_logfile;
	CoUninitialize();

	// Closing the window. Is it needed?
	MSG msg;
    ZeroMemory( &msg, sizeof(msg) );
    msg.message=WM_QUIT;
    TranslateMessage( &msg );
    DispatchMessage( &msg );
	
	// Remove textures
	RemoveTexture();

    if( m_pVB != NULL )
        m_pVB->Release();

    if( m_pd3dDevice != NULL )
        m_pd3dDevice->Release();

    if( m_pD3D != NULL )
        m_pD3D->Release();


    UnregisterClass( "D3D Filter", m_wc.hInstance );
}


//
// CreateInstance
//
// Provide the way for COM to create a CD3DFilter object
//
CUnknown * WINAPI CD3DFilter::CreateInstance(LPUNKNOWN punk, HRESULT *phr) {

    CD3DFilter *pNewObject = new CD3DFilter(NAME("D3D Filter"), punk, phr);
    if (pNewObject == NULL) {
        *phr = E_OUTOFMEMORY;
    }
    return pNewObject;

} // CreateInstance

//
// NonDelegatingQueryInterface
//
// Reveals IContrast and ISpecifyPropertyPages
//
STDMETHODIMP CD3DFilter::NonDelegatingQueryInterface(REFIID riid, void **ppv)
{
    CheckPointer(ppv,E_POINTER);

    if (riid == IID_ID3DFilter) {
        return GetInterface((ID3DFilter *) this, ppv);
    } else if (riid == IID_ISpecifyPropertyPages) {
        return GetInterface((ISpecifyPropertyPages *) this, ppv);
    } else {
        return CTransformFilter::NonDelegatingQueryInterface(riid, ppv);
    }

} // NonDelegatingQueryInterface


//
// Transform
//
// Copy the input sample into the output sample
// Then transform the output sample 'in place'
//
HRESULT CD3DFilter::Transform(IMediaSample *pIn, IMediaSample *pOut)
{
	*m_logfile << "Time between frames = " << (DXUtil_Timer(TIMER_GETABSOLUTETIME)-frameTime) << ";" << endl;

	FLOAT startTime;
	FLOAT stopTime;
	startTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);

	// if using processor
	/*
	HRESULT hr = UploadTexture(pIn);
	if(FAILED(hr))
		Msg(TEXT("UPLOAD TEXTURE FAILED"));
	stopTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
	*m_logfile << "Processor time = " << (stopTime-startTime) << ";" << endl;
	*/

	// if using graphics card
	// Upload frame to texture
	HRESULT hr = UploadTexture(pIn);
	if(FAILED(hr))
		Msg(TEXT("UPLOAD TEXTURE FAILED"));

	stopTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
	*m_logfile << "UploadTexture = " << (stopTime-startTime) << ";" << endl;

	startTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);

	// Render the texture in D3D
	Render();
	stopTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
	*m_logfile << "Render = " << (stopTime-startTime) << ";" << endl;

	startTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);

	// Download texture to pOut
	hr = DownloadTexture(pOut);
	if(FAILED(hr))
		Msg(TEXT("DOWNLOAD TEXTURE FAILED"));

	stopTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
	*m_logfile << "DownloadTexture = " << (stopTime-startTime) << ";" << endl;

	frameTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);

	m_logfile->flush();

	return S_OK;
} // Transform



//-----------------------------------------------------------------------------
// DownLoadTexture: A sample has been delivered. Copy the texture to output.
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::DownloadTexture( IMediaSample * pSample )
{    
	BYTE  *pBmpBuffer, *pTxtBuffer;     // Bitmap buffer, texture buffer
    LONG  lTxtPitch;                // Pitch of bitmap, texture
    
    // Get the video bitmap buffer
    pSample->GetPointer( &pBmpBuffer );

    // Lock the Texture

	D3DLOCKED_RECT d3dlr;
    if (FAILED(m_pRenderTargetSurface->LockRect(&d3dlr, 0, 0)))
        return E_FAIL;
   

    // Get the texture buffer & pitch
    pTxtBuffer = static_cast<byte *>(d3dlr.pBits);
    lTxtPitch = d3dlr.Pitch;
    
    // Copy the bits    
    for(int y = 0; y < m_lVidHeight; y++ ) {
        BYTE *pBmpBufferOld = pBmpBuffer;
        BYTE *pTxtBufferOld = pTxtBuffer;   
        for (int x = 0; x < m_lVidWidth; x++) {
            pBmpBuffer[0] = pTxtBuffer[0];
            pBmpBuffer[1] = pTxtBuffer[1];
            pBmpBuffer[2] = pTxtBuffer[2];
            pTxtBuffer += 4;
            pBmpBuffer += 3;
        }
        pBmpBuffer = pBmpBufferOld + m_lVidPitch;
        pTxtBuffer = pTxtBufferOld + lTxtPitch;
    }
        
    // Unlock the Texture
    if (FAILED(m_pRenderTargetSurface->UnlockRect()))
        return E_FAIL;

    return S_OK;
}

//-----------------------------------------------------------------------------
// UploadTexture: A sample has been delivered. Copy it to the texture.
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::UploadTexture( IMediaSample * pSample )
{
    BYTE  *pBmpBuffer, *pTxtBuffer;     // Bitmap buffer, texture buffer
    LONG  lTxtPitch;                // Pitch of bitmap, texture
    
    // Get the video bitmap buffer
    pSample->GetPointer( &pBmpBuffer );

    // Lock the Texture
    D3DLOCKED_RECT d3dlr;
    if (FAILED(m_pTexture->LockRect(0, &d3dlr, 0, 0)))
        return E_FAIL;
    
    // Get the texture buffer & pitch
    pTxtBuffer = static_cast<byte *>(d3dlr.pBits);
    lTxtPitch = d3dlr.Pitch;
    
    // Copy the bits    

    for(int y = 0; y < m_lVidHeight; y++ ) {
        BYTE *pBmpBufferOld = pBmpBuffer;
        BYTE *pTxtBufferOld = pTxtBuffer;   
        for (int x = 0; x < m_lVidWidth; x++) {
            pTxtBuffer[0] = pBmpBuffer[0];
            pTxtBuffer[1] = pBmpBuffer[1];
            pTxtBuffer[2] = pBmpBuffer[2];
			
			// Testing processor
            //pTxtBuffer[0] = (byte) (0.3*pBmpBuffer[0]+0.4*pBmpBuffer[1]+0.3*pBmpBuffer[2]);
            //pTxtBuffer[1] = (byte) (0.4*pBmpBuffer[0]+0.3*pBmpBuffer[1]+0.3*pBmpBuffer[2]);
            //pTxtBuffer[2] = (byte) (0.3*pBmpBuffer[0]+0.3*pBmpBuffer[1]+0.4*pBmpBuffer[2]);
			// Testing processor
            
			pTxtBuffer += 4;
            pBmpBuffer += 3;
        }
        pBmpBuffer = pBmpBufferOld + m_lVidPitch;
        pTxtBuffer = pTxtBufferOld + lTxtPitch;
    }

        
    // Unlock the Texture
    if (FAILED(m_pTexture->UnlockRect(0)))
        return E_FAIL;


	if (FAILED(m_pd3dDevice->SetPixelShaderConstantF(0, params, 1)))
    {
        Msg(TEXT("Could not set Pixel Shader params vector!"));
		return E_FAIL;
    }
    
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID CD3DFilter::Render()
{
	HRESULT hr;

	if(SUCCEEDED(m_pRenderToSurface->BeginScene(m_pRenderTargetSurface, NULL)))
    {
	//if(SUCCEEDED(m_pd3dDevice->SetRenderTarget(0,m_pRenderTargetSurface) ))
	//{
		// Clear the backbuffer and the zbuffer
		//if(FAILED(hr = m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0,255,255), 1.0f, 0 ))) {
		//	*m_logfile << LOCATION << "Failed to clear d3ddevice: " << (hr&0xffff) << endl;	
		//	m_logfile->flush();
		//}


		// Begin the scene
		m_pd3dDevice->BeginScene();

		// Setup the world, view, and projection matrices
		//SetupMatrices();

		// Set texture
		m_pd3dDevice->SetTexture( 0, m_pTexture );

		// Render the vertex buffer contents
		m_pd3dDevice->SetStreamSource( 0, m_pVB, NULL, sizeof(CUSTOMVERTEX) );
		m_pd3dDevice->SetVertexShader( m_hBasicVertexShader );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
		
		// Det var den hr j***a raden som var viktig fr att det skulle funka!!!
		m_pd3dDevice->SetFVF( D3DFVF_CUSTOMVERTEX );
		
		m_pd3dDevice->SetPixelShader( m_hBasicPixelShader );

		if(FAILED(hr = m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 ))) {
			if(hr==D3DERR_INVALIDCALL) {
				*m_logfile << LOCATION << "Failed to draw primitive (INVALID_CALL): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
			else {
				*m_logfile << LOCATION << "Failed to draw primitive (Unknown error): " << (hr&0xffff) << endl;	
				m_logfile->flush();
			}
		}

		FLOAT stopTime;
		FLOAT startTime;
		startTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
		//this->m_pClock->GetTime(&startTime);

		m_pRenderToSurface->EndScene(D3DX_FILTER_NONE);

		//this->m_pClock->GetTime(&stopTime);
		stopTime = DXUtil_Timer(TIMER_GETABSOLUTETIME);
		*m_logfile << "EndScene = " << (stopTime-startTime) << ";" << endl;

	}
    else
	{
		*m_logfile << LOCATION << "Failed in method render()!" << endl;	
		m_logfile->flush();
    }

}



//-----------------------------------------------------------------------------
// InitTexture: Creates all textures and surfaces 
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::InitTexture()
{
    HRESULT hr;
	// The texture which we copy the input to
	if( FAILED( hr = m_pd3dDevice->CreateTexture(
                    m_lVidWidth, m_lVidHeight,
                    1, 0, 
                    D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_pTexture, NULL ) ) )
    {
        Msg(TEXT("Could not create the D3DX texture!"));
        return hr;
    }

	// Rendering to surface
	if(FAILED(hr = D3DXCreateRenderToSurface(m_pd3dDevice, m_lVidWidth, m_lVidHeight, 
        D3DFMT_X8R8G8B8, FALSE, D3DFMT_UNKNOWN, &m_pRenderToSurface)))
    {
        Msg(TEXT("Could not create Renderer to surface object!"));
		return hr;
    }

	// The final output surface
	if(FAILED(hr = m_pd3dDevice->CreateOffscreenPlainSurface(m_lVidWidth, m_lVidHeight, 
        D3DFMT_X8R8G8B8, D3DPOOL_SCRATCH, &m_pRenderTargetSurface, NULL)))
    {
        Msg(TEXT("Could not create Surface object!"));
		return hr;
    }

	if (FAILED(hr =m_pd3dDevice->SetPixelShaderConstantF(0, params, 1)))
    {
        Msg(TEXT("Could not set Pixel Shader params vector!"));
		return hr;
    }

	// To avoid the texture sliding, half a texel must be added
	float displacement[4] = {0.01f,0.01f,0.0f,0.0f};
	/*displacement[0] = 0.01f/m_lVidWidth;
	displacement[1] = 0.01f/m_lVidHeight;
	displacement[2] = 0.0f;
	displacement[3] = 0.00f;*/
	//displacement[,0.01f,0.0f,0.0f};
	if (FAILED(hr =m_pd3dDevice->SetVertexShaderConstantF(0, displacement, 1)))                                      
    {
        Msg(TEXT("Could not set Vertex Shader displacement vector!"));
		return hr;
    }

	return S_OK;
}


//-----------------------------------------------------------------------------
// RemoveTexture: This method removes all textures and surfaces
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::RemoveTexture()
{
	if( m_pTexture != NULL )
        m_pTexture->Release();

	if( m_pRenderTargetSurface != NULL )
        m_pRenderTargetSurface->Release();

	if( m_pRenderToSurface != NULL )
        m_pRenderToSurface->Release();
	
	return S_OK;
}



//-----------------------------------------------------------------------------
// CheckMediaType: This method forces the graph to give us an R8G8B8 video
// type, making our copy to texture memory trivial.
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::CheckInputType(const CMediaType *pmt)
{
    HRESULT   hr = E_FAIL;
    VIDEOINFO *pvi;
    
    // Reject the connection if this is not a video type
    if( *pmt->FormatType() != FORMAT_VideoInfo ) {
        return E_INVALIDARG;
    }
    
    // Only accept RGB24
    pvi = (VIDEOINFO *)pmt->Format();
    if(IsEqualGUID( *pmt->Type(),    MEDIATYPE_Video)  &&
       IsEqualGUID( *pmt->Subtype(), MEDIASUBTYPE_RGB24))
    {
        hr = S_OK;
    }
    
    return hr;
}



//
// CheckTransform
//
// To be able to transform the formats must be identical
//
HRESULT CD3DFilter::CheckTransform(const CMediaType *mtIn,const CMediaType *mtOut)
{
    HRESULT hr;
    if (FAILED(hr = CheckInputType(mtIn))) {
	return hr;
    }

    // format must be a VIDEOINFOHEADER
    if (*mtOut->FormatType() != FORMAT_VideoInfo) {
	return E_INVALIDARG;
    }
    
    // formats must be big enough 
    if (mtIn->FormatLength() < sizeof(VIDEOINFOHEADER) ||
	mtOut->FormatLength() < sizeof(VIDEOINFOHEADER))
	return E_INVALIDARG;
    
    VIDEOINFO *pInput = (VIDEOINFO *) mtIn->Format();
	LONG lVidWidth  = pInput->bmiHeader.biWidth;
    LONG lVidHeight = abs(pInput->bmiHeader.biHeight);
	*m_logfile << "lVidWith = " << lVidWidth << ", lVidHeight = " << lVidHeight << endl;
    LONG lVidPitch = (lVidWidth * 3 + 3) & ~(3); // We are forcing RGB24
	

	if(lVidWidth!=m_lVidWidth || lVidHeight!=m_lVidHeight) {
		m_lVidWidth = lVidWidth;
		m_lVidHeight = lVidHeight;
		m_lVidPitch = lVidPitch;
		RemoveTexture();
		InitTexture();
	}
	
    VIDEOINFO *pOutput = (VIDEOINFO *) mtOut->Format();
    if (memcmp(&pInput->bmiHeader,&pOutput->bmiHeader,sizeof(BITMAPINFOHEADER)) == 0) {
	return NOERROR;
    }

    return E_INVALIDARG;
} // CheckTransform


//
// DecideBufferSize
//
// Tell the output pin's allocator what size buffers we
// require. Can only do this when the input is connected
//
HRESULT CD3DFilter::DecideBufferSize(IMemAllocator *pAlloc,ALLOCATOR_PROPERTIES *pProperties)
{
    // Is the input pin connected

    if (m_pInput->IsConnected() == FALSE) {
        return E_UNEXPECTED;
    }

    ASSERT(pAlloc);
    ASSERT(pProperties);
    HRESULT hr = NOERROR;

    pProperties->cBuffers = 1;
    pProperties->cbBuffer = m_pInput->CurrentMediaType().GetSampleSize();

    ASSERT(pProperties->cbBuffer);

    // If we don't have fixed sized samples we must guess some size

    if (!m_pInput->CurrentMediaType().bFixedSizeSamples) {
        if (pProperties->cbBuffer < 100000) {
            // nothing more than a guess!!
            pProperties->cbBuffer = 100000;
        }
    }

    // Ask the allocator to reserve us some sample memory, NOTE the function
    // can succeed (that is return NOERROR) but still not have allocated the
    // memory that we requested, so we must check we got whatever we wanted

    ALLOCATOR_PROPERTIES Actual;
    hr = pAlloc->SetProperties(pProperties,&Actual);
    if (FAILED(hr)) {
        return hr;
    }

    ASSERT( Actual.cBuffers == 1 );

    if (pProperties->cBuffers > Actual.cBuffers ||
            pProperties->cbBuffer > Actual.cbBuffer) {
                return E_FAIL;
    }
    return NOERROR;

} // DecideBufferSize


//
// GetMediaType
//
// Gives same mediatype to output
// as input
HRESULT CD3DFilter::GetMediaType(int iPosition, CMediaType *pMediaType)
{
    // Is the input pin connected

    if (m_pInput->IsConnected() == FALSE) {
        return E_UNEXPECTED;
    }

    // This should never happen

    if (iPosition < 0) {
        return E_INVALIDARG;
    }

    // Do we have more items to offer

    if (iPosition > 0) {
        return VFW_S_NO_MORE_ITEMS;
    }

	// The same media type for output
    *pMediaType = m_pInput->CurrentMediaType();
    return NOERROR;

} // GetMediaType


//
// setHWND
//
// Set the HWND
//
STDMETHODIMP CD3DFilter::setHWND(HWND *hwnd)
{
    return NOERROR;

} // get_betaLevel

STDMETHODIMP CD3DFilter::setParam1(float param)
{
	// Maybe should check that param is between 0 and 1?
	params[0] = param;
    return NOERROR;
} 

STDMETHODIMP CD3DFilter::setParam2(float param)
{
	params[1] = param;
    return NOERROR;
} 

STDMETHODIMP CD3DFilter::setParam3(float param)
{
	params[2] = param;
    return NOERROR;
} 

//
// GetPages
//
// This is the sole member of ISpecifyPropertyPages
// Returns the clsid's of the property pages we support
//
STDMETHODIMP CD3DFilter::GetPages(CAUUID *pPages)
{
    pPages->cElems = 1;
    pPages->pElems = (GUID *) CoTaskMemAlloc(sizeof(GUID));
    if (pPages->pElems == NULL) {
        return E_OUTOFMEMORY;
    }
    *(pPages->pElems) = CLSID_D3DFilterPropertyPage;
    return NOERROR;

} // GetPages


//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT CD3DFilter::InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( m_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Get the current desktop display mode, so we can set up a back
    // buffer of the same format
    D3DDISPLAYMODE d3ddm;
    if( FAILED( m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &d3ddm ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice. Since we are now
    // using more complex geometry, we will create a device with a zbuffer.
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = d3ddm.Format;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    // Create the D3DDevice
    if( FAILED( m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &m_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    // Turn off culling
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    return S_OK;
}



//
// DllRegisterServer
//
// Handle registration of this filter
//
STDAPI DllRegisterServer()
{
    return AMovieDllRegisterServer2( TRUE );

} // DllRegisterServer


//
// DllUnregisterServer
//
STDAPI DllUnregisterServer()
{
    return AMovieDllRegisterServer2( FALSE );

} // DllUnregisterServer


//-----------------------------------------------------------------------------
// Msg: Display an error message box if needed
//-----------------------------------------------------------------------------
void Msg(TCHAR *szFormat, ...)
{
    //TCHAR szBuffer[512];

    //va_list pArgs;
    //va_start(pArgs, szFormat);
    //_vstprintf(szBuffer, szFormat, pArgs);
    //va_end(pArgs);

    MessageBox(NULL, szFormat, TEXT("DirectShow Texture3D Sample"), 
               MB_OK | MB_ICONERROR);
}


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}